"""
Implementation of dFC methods.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import time

import numpy as np
from ksvd import ApproximateKSVD

from ..dfc import DFC
from ..time_series import TIME_SERIES
from .base_dfc_method import BaseDFCMethod

################################## Windowless ##################################

"""
by : https://github.com/nel215/ksvd

Reference: Rubinstein, R., Zibulevsky, M. and Elad, M., Efficient Implementation
of the K-SVD Algorithm using Batch Orthogonal Matching Pursuit Technical
Report - CS Technion, April 2008

Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signals.
    dt : float
        Sample spacing.

todo:
"""


class WINDOWLESS(BaseDFCMethod):

    def __init__(self, **params):
        self.logs_ = ""
        self.TPM = []
        self.FCS_ = []
        self.mean_act = []
        self.FCS_fit_time_ = None
        self.dFC_assess_time_ = None

        self.params_name_lst = [
            "measure_name",
            "is_state_based",
            "n_states",
            "normalization",
            "num_subj",
            "num_select_nodes",
            "num_time_point",
            "Fs_ratio",
            "noise_ratio",
            "num_realization",
            "session",
        ]
        self.params = {}
        for params_name in self.params_name_lst:
            if params_name in params:
                self.params[params_name] = params[params_name]
            else:
                self.params[params_name] = None

        self.params["measure_name"] = "Windowless"
        self.params["is_state_based"] = True

    @property
    def measure_name(self):
        return self.params["measure_name"]

    def estimate_FCS(self, time_series):

        assert (
            type(time_series) is TIME_SERIES
        ), "time_series must be of TIME_SERIES class."

        time_series = self.manipulate_time_series4FCS(time_series)

        # start timing
        tic = time.time()

        # time_series ~ gamma.dot(dictionary)
        self.aksvd = ApproximateKSVD(
            n_components=self.params["n_states"], transform_n_nonzero_coefs=1
        )
        self.dictionary = self.aksvd.fit(time_series.data.T).components_
        self.gamma = self.aksvd.transform(time_series.data.T)

        self.FCS_ = np.zeros(
            [self.params["n_states"], time_series.n_regions, time_series.n_regions]
        )
        for i in range(self.params["n_states"]):
            self.FCS_[i, :, :] = np.multiply(
                np.expand_dims(self.dictionary[i, :], axis=0).T,
                np.expand_dims(self.dictionary[i, :], axis=0),
            )

        self.Z = list()
        for i in range(self.gamma.shape[0]):
            self.Z.append(np.argwhere(self.gamma[i, :] != 0)[0, 0])

        # mean activation of states
        self.set_mean_activity(time_series)

        # record time
        self.set_FCS_fit_time(time.time() - tic)

        return self

    def estimate_dFC(self, time_series):

        assert (
            type(time_series) is TIME_SERIES
        ), "time_series must be of TIME_SERIES class."

        assert (
            len(time_series.subj_id_lst) == 1
        ), "this function takes only one subject as input."

        time_series = self.manipulate_time_series4dFC(time_series)

        # start timing
        tic = time.time()

        gamma = self.aksvd.transform(time_series.data.T)

        Z = list()
        for i in range(time_series.n_time):
            Z.append(np.argwhere(gamma[i, :] != 0)[0, 0])

        # record time
        self.set_dFC_assess_time(time.time() - tic)

        dFC = DFC(measure=self)
        dFC.set_dFC(FCSs=self.FCS_, FCS_idx=Z, TS_info=time_series.info_dict)
        return dFC


##########################################################################################



"""
Implementation of dFC methods.
the parent dFC method class

Created on Jun 29 2023
@author: Mohammad Torabi
"""

from copy import deepcopy

import numpy as np

from ..dfc_utils import SW_downsample, visualize_FCS

################################# BaseDFCMethod class ####################################

"""
todo:
- type annotation
"""


class BaseDFCMethod:

    TF_methods_name_lst = ["CWT_mag", "CWT_phase_r", "CWT_phase_a", "WTC"]

    sw_methods_name_lst = [
        "pear_corr",
        "MI",
        "GraphLasso",
    ]

    base_methods_name_lst = ["SlidingWindow", "Time-Freq"]

    def __init__(self):
        self.measure_name = ""
        self.is_state_based = bool()
        self._stat = []
        self.TPM = []
        self.params = {}
        self.TS_info_ = {}
        self.FCS_fit_time_ = None
        self.dFC_assess_time_ = None
        self.logs_ = ""

    @property
    def FCS_fit_time(self):
        return self.FCS_fit_time_

    @property
    def dFC_assess_time(self):
        return self.dFC_assess_time_

    @property
    def TS_info(self):
        # info of the time series used to train/estimate FCSs
        return self.TS_info_

    @property
    def is_state_based(self):
        return self.params["is_state_based"]

    @property
    def FCS(self):
        return self.FCS_

    # test
    @property
    def FCS_dict(self):
        # returns a dict including FCS matrices

        if not self.is_state_based:
            return None

        C_A = self.FCS
        FCSs = {}
        for k in range(C_A.shape[0]):
            FCSs["FCS" + str(k + 1)] = C_A[k, :, :]

        return FCSs

    @property
    def info(self):
        return self.params

    @property
    def logs(self):
        print(self.logs_)

    def issame(self, dFC):
        if type(self) == type(dFC):
            for param_name in self.params:
                if self.params[param_name] != dFC.params[param_name]:
                    return False
        else:
            return False
        return True

    # test
    def param_match(self, **param_dict):
        for param in param_dict:
            if param in self.params:
                if type(param_dict[param]) is list:
                    if not self.params[param] in param_dict[param]:
                        return False
                else:
                    if self.params[param] != param_dict[param]:
                        return False
        return True

    def set_FCS_fit_time(self, time):
        self.FCS_fit_time_ = time

    def set_dFC_assess_time(self, time):
        self.dFC_assess_time_ = time

    def set_mean_activity(self, time_series):
        # mean activity of regions at each state
        if self.is_state_based:
            if "sw_method" in self.params_name_lst:
                SUBJECTs = time_series.subj_id_lst
                TS_data = None
                for subject in SUBJECTs:
                    subj_TS = time_series.get_subj_ts(subjs_id=subject).data
                    new_TS_data = SW_downsample(
                        data=subj_TS.T,
                        Fs=time_series.Fs,
                        W=self.params["W"],
                        n_overlap=self.params["n_overlap"],
                        tapered_window=self.params["tapered_window"],
                    ).T
                    if TS_data is None:
                        TS_data = new_TS_data
                    else:
                        TS_data = np.concatenate((TS_data, new_TS_data), axis=1)
            else:
                TS_data = time_series.data
            mean_act = list()
            for i in np.unique(self.Z):
                ids = np.array([int(state == i) for state in self.Z])
                mean_act.append(np.average(TS_data, weights=ids, axis=1))
            self.mean_act = np.array(mean_act)
        else:
            self.mean_act = None

    def estimate_FCS(self, time_series=None):
        pass

    def estimate_dFC(self, time_series=None):
        pass

    def manipulate_time_series4FCS(self, time_series):
        """
        passing None to params will not change the time series
        num_realization is not implemented yet
        """

        new_time_series = deepcopy(time_series)

        # SUBJECTs
        if not self.params["num_subj"] is None:
            new_time_series.select_subjs(num_subj=self.params["num_subj"])
        # SPATIAL RESOLUTION
        if not self.params["num_select_nodes"] is None:
            new_time_series.spatial_downsample(
                num_select_nodes=self.params["num_select_nodes"], rand_node_slct=False
            )
        # TEMPORAL RESOLUTION
        if not self.params["Fs_ratio"] is None:
            new_time_series.Fs_resample(Fs_ratio=self.params["Fs_ratio"])
        # NORMALIZE
        if self.params["normalization"]:
            new_time_series.normalize()
        # NOISE
        if not self.params["noise_ratio"] is None:
            new_time_series.add_noise(
                noise_ratio=self.params["noise_ratio"], mean_noise=0
            )
        # NUMBER OF TIME POINTS
        if not self.params["num_time_point"] is None:
            new_time_series.truncate(
                start_point=0, end_point=self.params["num_time_point"] - 1
            )

        self.TS_info_ = new_time_series.info_dict

        return new_time_series

    def manipulate_time_series4dFC(self, time_series):
        """
        passing None to params will not change the time series
        num_realization is not implemented yet
        """

        new_time_series = deepcopy(time_series)

        # SPATIAL RESOLUTION
        if not self.params["num_select_nodes"] is None:
            new_time_series.spatial_downsample(
                num_select_nodes=self.params["num_select_nodes"], rand_node_slct=False
            )
        # TEMPORAL RESOLUTION
        if not self.params["Fs_ratio"] is None:
            new_time_series.Fs_resample(Fs_ratio=self.params["Fs_ratio"])
        # NORMALIZE
        if self.params["normalization"]:
            new_time_series.normalize()
        # NOISE
        if not self.params["noise_ratio"] is None:
            new_time_series.add_noise(
                noise_ratio=self.params["noise_ratio"], mean_noise=0
            )
        # NUMBER OF TIME POINTS
        if not self.params["num_time_point"] is None:
            new_time_series.truncate(
                start_point=0, end_point=self.params["num_time_point"] - 1
            )

        return new_time_series

    def visualize_states(self):
        pass

    # todo : use FCS_dict func in this func
    def visualize_FCS(
        self, normalize=True, fix_lim=True, save_image=False, output_root=None
    ):

        visualize_FCS(
            self,
            normalize=normalize,
            fix_lim=fix_lim,
            save_image=save_image,
            output_root=output_root,
        )


################################## NEW METHOD ##################################

"""
by : web link

Reference: ##

Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signals.
    dt : float
        Sample spacing.

todo:

import needed_toolbox

class method_name(dFC):

    def __init__(self, **params):
        self.FCS_ = []
        self.logs_ = ''

        self.params_name_lst = ['measure_name', 'is_state_based', 'n_states',
            'normalization', 'num_subj', 'num_select_nodes', 'num_time_point',
            'Fs_ratio', 'noise_ratio', 'num_realization', 'session']
        self.params = {}
        for params_name in self.params_name_lst:
            if params_name in params:
                self.params[params_name] = params[params_name]
            else:
                self.params[params_name] = None

        self.params['specific_param'] = value
        self.params['measure_name'] = 'method_name'
        self.params['is_state_based'] = True/False

    @property
    def measure_name(self):
        return self.params['measure_name']

    def estimate_FCS(self, time_series):

        assert type(time_series) is TIME_SERIES, \
            "time_series must be of TIME_SERIES class."

        time_series = self.manipulate_time_series4FCS(time_series)

        # start timing
        tic = time.time()

        # calc FCSs

        # calc self.Z

        # mean activation of states
        self.set_mean_activity(time_series)

        # record time
        self.set_FCS_fit_time(time.time() - tic)

        return self

    def estimate_dFC(self, time_series):

        assert type(time_series) is TIME_SERIES, \
            "time_series must be of TIME_SERIES class."

        assert len(time_series.subj_id_lst)==1, \
            'this function takes only one subject as input.'

        time_series = self.manipulate_time_series4dFC(time_series)

        # start timing
        tic = time.time()

        # calc FCSs and FCS_idx

        # record time
        self.set_dFC_assess_time(time.time() - tic)

        dFC = DFC(measure=self)
        dFC.set_dFC(FCSs=self.FCS_, FCS_idx=FCS_idx, TS_info=time_series.info_dict)
        return dFC
"""



# -*- coding: utf-8 -*-
"""
Analytical functions for the comparison framework.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

from copy import deepcopy

import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy import stats
from statsmodels.formula.api import ols

from ..dfc_utils import dFC_mat2vec, mat_reorder, zip_name

################################# Analytical Functions ####################################

############## STAT Functions ##############


def make_sim_distribution(sim_mats_lst, name_lst, zip_names=True):
    """
    each sim_mat in the sim_mat_lst corresponds
    to a subj. the name_lst must correspond to the
    columns and rows of sim_mats
    """
    output = {}
    for sim_mat in sim_mats_lst:

        for i, name_i in enumerate(name_lst):
            for j, name_j in enumerate(name_lst):

                if j >= i:
                    continue

                if zip_names:
                    name_i_used = zip_name(name_i)
                    name_j_used = zip_name(name_j)
                else:
                    name_i_used = name_i
                    name_j_used = name_j

                if not name_i_used in output:
                    output[name_i_used] = {}
                if not name_j_used in output[name_i_used]:
                    output[name_i_used][name_j_used] = {"sim": list(), "": list()}

                output[name_i_used][name_j_used]["sim"].append(sim_mat[i, j])
                output[name_i_used][name_j_used][""].append("sim")
    return output


def two_way_anova(data):
    """
    perform two-way anova
    target: sim
    factor1: session
    factor2: direction
    """
    df = pd.DataFrame(data)

    # Performing two-way ANOVA
    model = ols(
        "sim ~ C(session) + C(direction) +\
    C(session):C(direction)",
        data=df,
    ).fit()

    return sm.stats.anova_lm(model, type=2)


def convert_pvalue_to_asterisks(pvalue):
    if pvalue <= 0.0001:
        return "****"
    elif pvalue <= 0.001:
        return "***"
    elif pvalue <= 0.01:
        return "**"
    elif pvalue <= 0.05:
        return "*"
    return "ns"


############## Randomization Functions ##############


def randomize_time(dFC_dict, N):
    """ """
    output = {}
    for n in range(N):

        for i, measure_i_name in enumerate(dFC_dict):

            dFC_mat_i = dFC_dict[measure_i_name]

            # randomize the temporal order
            n_time = dFC_mat_i.shape[0]
            idx = np.random.choice(n_time, n_time, replace=False)
            dFC_mat_i = dFC_mat_i[idx, :, :]

            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i)

            for j, measure_j_name in enumerate(dFC_dict):

                if j > i:
                    continue
                if not measure_i_name in output:
                    output[measure_i_name] = {}
                if not measure_j_name in output[measure_i_name]:
                    output[measure_i_name][measure_j_name] = {"sim": list(), "": list()}

                dFC_mat_j = dFC_dict[measure_j_name]

                # randomize the temporal order
                n_time = dFC_mat_j.shape[0]
                idx = np.random.choice(n_time, n_time, replace=False)
                dFC_mat_j = dFC_mat_j[idx, :, :]

                dFC_mat_j_vec = dFC_mat2vec(dFC_mat_j)

                sim, p = stats.spearmanr(dFC_mat_i_vec.flatten(), dFC_mat_j_vec.flatten())
                output[measure_i_name][measure_j_name]["sim"].append(sim)
                output[measure_i_name][measure_j_name][""].append("sim")

    return output


def suffle_dFC(dFC_mat, mode):
    """
    dFC_mat = ndarray(time, region, region)
    mode can be 'temporal', 'spatial',
    or 'all'
    """
    new_dFC_mat = deepcopy(dFC_mat)
    if mode == "temporal":
        n_time = new_dFC_mat.shape[0]
        new_order = np.random.choice(n_time, n_time, replace=False)
        new_dFC_mat = new_dFC_mat[new_order, :, :]
    elif mode == "spatial":
        n_region = new_dFC_mat.shape[1]
        new_order = np.random.choice(n_region, n_region, replace=False)
        for k, mat in enumerate(new_dFC_mat):
            new_dFC_mat[k, :, :] = mat_reorder(new_dFC_mat[k, :, :], new_order)
    elif mode == "all":
        # spatial
        n_region = new_dFC_mat.shape[1]
        new_order_regions = np.random.choice(n_region, n_region, replace=False)
        for k, mat in enumerate(new_dFC_mat):
            new_dFC_mat[k, :, :] = mat_reorder(new_dFC_mat[k, :, :], new_order_regions)
        # temporal
        n_time = new_dFC_mat.shape[0]
        new_order_time = np.random.choice(n_time, n_time, replace=False)
        new_dFC_mat = new_dFC_mat[new_order_time, :, :]

    return new_dFC_mat


def randomized_dFC_sim(dFC_dict, N, mode):
    """
    mode can be 'temporal', 'spatial',
    or 'all'
    'spatial': this will result in different methods having
    different spatial/region orders but still the
    same temporal order
    'temporal': this will result in different methods having
    different temporal orders but still the
    same spatial order
    'all': this will result in different methods having
    different temporal orders AND different spatial order
    """
    output = {}
    for n in range(N):

        for i, measure_i_name in enumerate(dFC_dict):

            dFC_mat_i = dFC_dict[measure_i_name]

            # randomize the spatial (regions) order
            dFC_mat_i = suffle_dFC(dFC_mat_i, mode=mode)

            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i)

            for j, measure_j_name in enumerate(dFC_dict):

                if j > i:
                    continue
                if not measure_i_name in output:
                    output[measure_i_name] = {}
                if not measure_j_name in output[measure_i_name]:
                    output[measure_i_name][measure_j_name] = {"sim": list(), "": list()}

                dFC_mat_j = dFC_dict[measure_j_name]

                # randomize the temporal order
                dFC_mat_j = suffle_dFC(dFC_mat_j, mode=mode)

                dFC_mat_j_vec = dFC_mat2vec(dFC_mat_j)

                sim, p = stats.spearmanr(dFC_mat_i_vec.flatten(), dFC_mat_j_vec.flatten())
                output[measure_i_name][measure_j_name]["sim"].append(sim)
                output[measure_i_name][measure_j_name][""].append("sim")

    return output


def dFC_rand_generator(FCS, n_time):
    """
    generate a dFC mat of length n_time
    using spatial FC patterns in FCS = (num_pattern, ROI, ROI)
    """
    dFC_rand = None
    idx = np.random.choice(FCS.shape[0], n_time, replace=True)
    dFC_rand = FCS[idx, :, :]
    return dFC_rand


def dFC_rand_sim(FCS_dict, n_time, N):
    """
    for random state TC similarity assessment
    """
    output = {}
    for n in range(N):

        for i, measure_i_name in enumerate(FCS_dict):

            dFC_rand = dFC_rand_generator(FCS_dict[measure_i_name], n_time=n_time)
            dFC_mat_i = dFC_rand
            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i)

            for j, measure_j_name in enumerate(FCS_dict):

                if j > i:
                    continue
                if not measure_i_name in output:
                    output[measure_i_name] = {}
                if not measure_j_name in output[measure_i_name]:
                    output[measure_i_name][measure_j_name] = {"sim": list(), "": list()}

                dFC_rand = dFC_rand_generator(FCS_dict[measure_j_name], n_time=n_time)
                dFC_mat_j = dFC_rand
                dFC_mat_j_vec = dFC_mat2vec(dFC_mat_j)

                sim, p = stats.spearmanr(dFC_mat_i_vec.flatten(), dFC_mat_j_vec.flatten())
                output[measure_i_name][measure_j_name]["sim"].append(sim)
                output[measure_i_name][measure_j_name][""].append("sim")
    return output


############## Hierarchical Clustering ##############


def correct_order(s):
    list = s.split("-")
    list = [int(item) for item in list]
    list.sort()
    return "-".join(str(x) for x in list)


def open_trees(Z, num_leaf):
    """
    replace trees in Z by their leaves
    """
    Z_copy = deepcopy(Z)
    Z_new = []
    for tree in Z_copy:
        Z_new.append([tree[0], tree[1], tree[2], tree[3]])
    encode_dict = {}
    counter = num_leaf
    for tree in Z_new:
        if tree[0] >= num_leaf:
            tree[0] = encode_dict[tree[0]]
        else:
            tree[0] = str(int(tree[0]))
        if tree[1] >= num_leaf:
            tree[1] = encode_dict[tree[1]]
        else:
            tree[1] = str(int(tree[1]))
        encode_dict[counter] = tree[0] + "-" + tree[1]
        encode_dict[counter] = correct_order(encode_dict[counter])
        counter += 1
    return Z_new


def is_trees_equal(trees_1, trees_2):
    """
    trees_2 is the reference
    """
    for tree in trees_1:
        if (not [tree[0], tree[1]] in trees_2) and (not [tree[1], tree[0]] in trees_2):
            return False
    return True


def is_in_Z_clstrs(trees, Z_clstrs, trees_key):
    for key in Z_clstrs:
        if is_trees_equal(trees, Z_clstrs[key][trees_key]):
            return key
    return None


def cluster_Z(Z_lst, num_leaf):
    """
    Z_lst is the list of linkages of samples
    num_leaf is the number of objects in clustering
    """
    Z_clstrs = {}
    counter = 0
    for Z in Z_lst:
        # replace trees in Z by their leaves
        Z_open = open_trees(Z, num_leaf)
        trees = [[tree[0], tree[1]] for tree in Z_open]
        distances = [tree[2] for tree in Z]
        clstr_idx = is_in_Z_clstrs(trees, Z_clstrs, trees_key="trees")

        if clstr_idx is None:
            Z_clstrs[counter] = {
                "Z": Z,
                "trees": trees,
                "freq": 1,
                "distance_lst": [distances],
            }
            counter += 1
        else:
            Z_clstrs[clstr_idx]["freq"] += 1
            Z_clstrs[clstr_idx]["distance_lst"].append(distances)
    return Z_clstrs



import os
import time

import hdf5storage
import numpy as np
import scipy.io as sio
from functions.dFC_funcs import *

os.environ["MKL_NUM_THREADS"] = "64"
os.environ["NUMEXPR_NUM_THREADS"] = "64"
os.environ["OMP_NUM_THREADS"] = "64"

print(
    "################################# subject-level similarity measurement CODE started running ... #################################"
)

################################# Parameters #################################
###### DATA PARAMETERS ######

input_root = "./"
output_root = "./"

################################# LOAD #################################

dFC_analyzer = np.load(input_root + "dFC_analyzer.npy", allow_pickle="TRUE").item()
data_loader = np.load(input_root + "data_loader.npy", allow_pickle="TRUE").item()

################################# LOAD ASSESSED dFC #################################

task_id = int(os.getenv("SGE_TASK_ID"))
subj_id = data_loader.SUBJECTS[task_id - 1]  # SGE_TASK_ID starts from 1 not 0
folder = input_root + "dFC_assessed/SUBJ_" + str(subj_id)

ALL_RECORDS = os.listdir(folder + "/")
ALL_RECORDS = [i for i in ALL_RECORDS if "dFCM" in i]
ALL_RECORDS.sort()
dFCM_lst = list()
for s in ALL_RECORDS:
    dFCM = np.load(folder + "/" + s, allow_pickle="TRUE").item()
    dFCM_lst.append(dFCM)
print("assessed dFCMs loaded ...")

################################# SIMILARITY MEASUREMENT #################################

similarity_assessment = SIMILARITY_ASSESSMENT(dFCM_lst=dFCM_lst)

tic = time.time()
print("Measurement Started ...")

print("Similarity measurement started...")
SUBJ_output = similarity_assessment.run(
    FILTERS=dFC_analyzer.hyper_param_info, downsampling_method="default"
)
print("Similarity measurement done.")

print(f"Measurement required {time.time() - tic:0.3f} seconds.")

# Save
folder = output_root + "similarity_measured"
if not os.path.exists(folder):
    os.makedirs(folder)

np.save(folder + "/SUBJ_" + str(subj_id) + "_output.npy", SUBJ_output)
#################################################################################
